﻿
Imports Microsoft.DirectX
Imports Microsoft.DirectX.Direct3D

Module Pick

    Private hit As Boolean
    Private triFaceid As Int32
    Private a As Single
    Private b As Single
    Private dist As Single

    Private vIn As Vector3
    Private vNear As Vector3
    Private vFar As Vector3
    Private vDir As Vector3

    Private PickedPoint_Global As Vector3
    Private PickedPoint_Local As Vector3
    Private PickedObjIndex As Int32



    Friend Sub Pick_FromViewport(ByVal X As Int32, _
                                 ByVal Y As Int32, _
                                 Optional ByVal ExcludedObjectIndex As Int32 = -1)

        Dim vport As Viewport
        Dim proj As Matrix
        Dim view As Matrix

        vport = DX_Device.Viewport

        view = mtx_CameraView
        proj = mtx_CameraProj

        PickedObjIndex = -1
        dist = 1.0E+21

        For i As Int32 = 0 To UBound(Objects_3D)

            If i = ExcludedObjectIndex Then Continue For

            With Objects_3D(i)

                'If Not .IsMovingObject Then Continue For

                If .MeshX.msh Is Nothing Then Continue For

                ' ---------------------------------------------- compute point on Near Clip plane at cursor
                vIn.X = X
                vIn.Y = Y
                vIn.Z = 0
                vNear = Vector3.Unproject(vIn, vport, proj, view, .MeshX.mtx)

                ' ---------------------------------------------- compute point on far clip plane at cursor
                vIn.Z = 1
                vFar = Vector3.Unproject(vIn, vport, proj, view, .MeshX.mtx)

                ' ---------------------------------------------- compute direction vector
                vDir = Vector3.Subtract(vFar, vNear)

                Dim closest As IntersectInformation

                If .MeshX.msh.Intersect(vNear, vDir, closest) Then
                    ' ------------------------------ this comparation can fail if farplane / nearplane > 100k
                    If closest.Dist <= dist Then
                        PickedObjIndex = i
                        triFaceid = closest.FaceIndex
                        a = closest.U
                        b = closest.V
                        dist = closest.Dist
                        PickedPoint_Global = Camera.Pos + vDir * dist
                        PickedPoint_Local = Vector3.TransformCoordinate(PickedPoint_Global, .MeshX.mtx)
                    End If
                End If
            End With
        Next
    End Sub



    Friend Sub RayPick(ByRef vOrig As Vector3, ByRef vDir As Vector3, Optional ByVal ExcludedObjectIndex As Int32 = -1)

        PickedObjIndex = -1
        dist = 1.0E+21

        vNear = vOrig

        For i As Int32 = 0 To UBound(Objects_3D)

            If i = ExcludedObjectIndex Then Continue For

            With Objects_3D(i)
                Dim closest As IntersectInformation

                If .MeshX.msh Is Nothing Then Continue For

                If .MeshX.msh.Intersect(vNear, vDir, closest) Then
                    If closest.Dist < dist Then
                        PickedObjIndex = i
                        triFaceid = closest.FaceIndex
                        a = closest.U
                        b = closest.V
                        dist = closest.Dist
                        PickedPoint_Global = Camera.Pos + vDir * dist
                        PickedPoint_Local = Vector3.TransformCoordinate(PickedPoint_Global, .MeshX.mtx)
                    End If
                End If
            End With
        Next
    End Sub



    Friend Function ViewportPick_GetRayTo(ByVal x As Int32, ByVal y As Int32) As Vector3

        Dim rayForward As Vector3 = Camera.Dir
        Dim rayFrom As Vector3 = Camera.Pos
        Dim lookAt As Vector3 = Camera.At

        rayForward.Normalize()
        rayForward = (rayForward * f_FarPlane)

        Dim vertical As Vector3 = VEC3(0, 1, 0)
        Dim horizontal As Vector3

        horizontal = Vector3.Cross(rayForward, VEC3(vertical.X * -1, vertical.Y * -1, vertical.Z * -1))
        horizontal.Normalize()
        vertical = Vector3.Cross(horizontal, VEC3(rayForward.X * -1, rayForward.Y * -1, rayForward.Z * -1))
        vertical.Normalize()

        Dim FOV As Single = 50 / Camera.Lens
        Dim tanfov As Single = CSng(Math.Tan(0.5 * FOV))

        horizontal *= 2.0F * f_FarPlane * tanfov
        vertical *= 2.0F * f_FarPlane * tanfov

        If (f_ViewAspect >= 1) Then
            horizontal.Scale(f_ViewAspect)
        Else
            vertical.Scale(1 / f_ViewAspect)
        End If

        Dim rayToCenter As Vector3 = rayFrom + rayForward

        Dim dHor As Vector3 = horizontal * (1.0F / DX_Wnd.Width)
        Dim dVert As Vector3 = vertical * (1.0F / DX_Wnd.Height)

        Dim rayTo As Vector3 = rayToCenter - 0.5F * horizontal + 0.5F * vertical

        rayTo += dHor * x
        rayTo -= dVert * y

        Return rayTo
    End Function



    Private Structure FaceIndexInt16
        Dim v1 As Short
        Dim v2 As Short
        Dim v3 As Short
    End Structure

    Private Structure FaceIndexInt32
        Dim v1 As Int32
        Dim v2 As Int32
        Dim v3 As Int32
    End Structure

    Private Function IntegerToLong(ByRef n As Short) As Int32
        IntegerToLong = n And &HFFFF
    End Function

    Private Sub GetVertsIndexesFromFace(ByVal m As Mesh, _
                                    ByVal FaceIndex As Int32, _
                                    ByRef iv0 As Int32, ByRef iv1 As Int32, ByRef iv2 As Int32)

        Dim gs As GraphicsStream

        If m.IndexBuffer.Description.Is16BitIndices Then
            Dim aInt16 As FaceIndexInt16
            gs = m.IndexBuffer.Lock(FaceIndex * 6, 6, LockFlags.ReadOnly)
            aInt16 = CType(gs.Read(GetType(FaceIndexInt16)), FaceIndexInt16)
            m.IndexBuffer.Unlock()

            iv0 = IntegerToLong(aInt16.v1) '<-- convert to int to avoid negative numbers
            iv1 = IntegerToLong(aInt16.v2) '<-- convert to int to avoid negative numbers
            iv2 = IntegerToLong(aInt16.v3) '<-- convert to int to avoid negative numbers
        Else
            Dim aInt32 As FaceIndexInt32
            gs = m.IndexBuffer.Lock(FaceIndex * 12, 12, LockFlags.ReadOnly)
            aInt32 = CType(gs.Read(GetType(FaceIndexInt32)), FaceIndexInt32)
            m.IndexBuffer.Unlock()

            iv0 = aInt32.v1
            iv1 = aInt32.v2
            iv2 = aInt32.v3
        End If
        'Debug.Print(iv1.ToString & vbTab & iv2.ToString & vbTab & iv3.ToString)
    End Sub



    ' ===============================================================================================
    '   PUBLIC FUNCTIONS
    ' ===============================================================================================

    Friend Function ViewportPick_GetPickedPoint_Global() As Vector3
        Return PickedPoint_Global
    End Function

    Friend Function ViewportPick_GetPickedPoint_Local() As Vector3
        Return PickedPoint_Local
    End Function

    Friend Function Pick_GetPickedObjIndex() As Int32
        Return PickedObjIndex
    End Function

    Friend Function ViewportPick_GetTriFaceId() As Int32
        Return triFaceid
    End Function


    Friend Function ViewportPick_GetVertexId() As Int32
        Dim iv0, iv1, iv2, index As Int32
        GetVertsIndexesFromFace(Objects_3D(PickedObjIndex).MeshX.msh, triFaceid, iv0, iv1, iv2)

        'The barycentric coordinate of the third vertex (vertex 2) of the triangle is 1 - U - V
        Dim c As Single = 1 - a - b

        If a > c And a > b Then
            index = iv1
        Else
            If c > a And c > b Then
                index = iv0
            Else
                index = iv2
            End If
        End If

        'Debug.Print(a.ToString("0.0   ") & b.ToString("0.0   ") & c.ToString("0.0   "))
        'Debug.Print(index.ToString & vbCr & vbCr)

        Return index
    End Function


End Module
