#pragma once

#include "DXUTcamera.h"

class Terrain;

class Camera : public CFirstPersonCamera
{
public:
	Camera(void);
	~Camera(void);

	void FrameMove( float elapsedTime );	
   
	bool FrustumSphereCheck( D3DXVECTOR3 &center, float radius );
	// Cube check using box center and size
	bool FrustumBoxCheck( D3DXVECTOR3 &center, float size );
	// Cube cheek using min and max extent of the box
	bool FrustumBoxCheck( D3DXVECTOR3 &min, D3DXVECTOR3 &max );	

	void CalculateFrustumCorners();	
    void CalculateFrustumPlanes();

	void SetPosition( D3DXVECTOR3 v )	{ m_vEye = v; }
	const D3DXVECTOR3* GetEyePt() const { return (D3DXVECTOR3*)&m_vEye; }
	
	const float GetYawAngle()			{ return m_fCameraYawAngle; }
	const float GetPitchAngle()			{ return m_fCameraPitchAngle; }
	D3DXVECTOR3 *GetDirection()			{ return &(m_vLookAt - m_vEye); }
	const float GetFov()				{ return m_fFOV; }
	const float GetAspect()				{ return m_fAspect; }		

	D3DXVECTOR3 *GetFrustumPoints()		{ return mFrustumPoints; }

protected:

	// Camera planes to store view frustum
	D3DXPLANE mPlanes[6];	
	D3DXVECTOR3 mFrustumPoints[8];
};

// A spring based chase camera
class ChaseCamera : public Camera
{
public:
	ChaseCamera();

	virtual void FrameMove( FLOAT elapsedTime );	
	virtual void Reset();
	void SetChasePosition( D3DXVECTOR3 &pos ) { mChasePosition = pos; }
	void SetChaseDirection( D3DXVECTOR3 &dir ) { mChaseDirection = dir; }
	void UpdateWorldPositions();

private:
	float mStiffness;
	float mDamping;
	float mMass;
	D3DXVECTOR3 mTargetPosition;
	D3DXVECTOR3 mTargetPositionOffset;
	D3DXVECTOR3 mChaseDirection;
	D3DXVECTOR3 mChasePosition;
	D3DXVECTOR3 mLookAtOffset;
};

